
  source("dataS/scripts/vehicles/specializations/TrailerToggleTipEvent.lua");
  source("dataS/scripts/vehicles/specializations/TrailerTipAreaEvent.lua");
  
  Abschieber = {};
  
  Trailer.TIPSTATE_CLOSED = 0;
  Trailer.TIPSTATE_OPENING = 1;
  Trailer.TIPSTATE_OPEN = 2;
  Trailer.TIPSTATE_CLOSING = 3;
  
  function Abschieber.prerequisitesPresent(specializations)
      return SpecializationUtil.hasSpecialization(Fillable, specializations);
  end;
  
  function Abschieber:load(xmlFile)
  
      self.toggleTipState = SpecializationUtil.callSpecializationsFunction("toggleTipState");
      self.onStartTip = SpecializationUtil.callSpecializationsFunction("onStartTip");
      self.onEndTip = SpecializationUtil.callSpecializationsFunction("onEndTip");
      self.enableTipAnimation = SpecializationUtil.callSpecializationsFunction("enableTipAnimation");
      self.disableCurrentTipAnimation = SpecializationUtil.callSpecializationsFunction("disableCurrentTipAnimation");
      self.setCurrentTipAnimationTime = SpecializationUtil.callSpecializationsFunction("setCurrentTipAnimationTime");
      self.getCurrentTipAnimationTime = Abschieber.getCurrentTipAnimationTime;
      self.getCurrentTipAnimationDuration = Abschieber.getCurrentTipAnimationDuration;
      self.setFillLevelUpload = Utils.prependedFunction(self.setFillLevelUpload, Abschieber.setFillLevelUpload);
	  self.setFillLevelReset = Utils.prependedFunction(self.setFillLevelReset, Abschieber.setFillLevelReset);
      self.getCurrentFruitType = Abschieber.getCurrentFruitType;
      self.updateTipping = Abschieber.updateTipping;
	  self.resetTipping = Abschieber.resetTipping;
      self.getCanTipToGround = Abschieber.getCanTipToGround;
      self.getCanTip = Abschieber.getCanTip;
	  -- self.getAttachedTrailersFillLevelAndCapacity = Utils.overwrittenFunction(self.getAttachedTrailersFillLevelAndCapacity, Abschieber.getAttachedTrailersFillLevelAndCapacity)
	  -- self.orgCapacity = self.capacity
      self.tipText = Utils.getXMLI18NValue(xmlFile, "vehicle", getXMLString, "tipText", g_i18n:getText("Dump"), self.customEnvironment, false);
  
  
      self.lastFillDelta = 0;
  
      self.tipAnimations = {};
      local i = 0;
      while true do
          local key = string.format("vehicle.tipAnimations.tipAnimation(%d)", i);
          if not hasXMLProperty(xmlFile, key) then
              break;
          end
  
          local tipAnimation = {};
          tipAnimation.dischargeEndTime = getXMLFloat(xmlFile, key.."#dischargeEndTime");
          if tipAnimation.dischargeEndTime ~= nil then
              tipAnimation.dischargeEndTime = tipAnimation.dischargeEndTime * 1000;
          end
          tipAnimation.dischargeStartTime = Utils.getNoNil(getXMLFloat(xmlFile, key.."#dischargeStartTime"), 0)*1000;
          local animationName = getXMLString(xmlFile, key.."#animationName");
  
          tipAnimation.animationOpenSpeedScale = getXMLFloat(xmlFile, key.."#openSpeedScale");
          tipAnimation.animationCloseSpeedScale = getXMLFloat(xmlFile, key.."#closeSpeedScale");
  
          local speedScale = Utils.getNoNil(getXMLFloat(xmlFile, key.."#speedScale"), 1);
          if tipAnimation.animationOpenSpeedScale == nil then
              tipAnimation.animationOpenSpeedScale = speedScale;
          end
          if tipAnimation.animationCloseSpeedScale == nil then
              tipAnimation.animationCloseSpeedScale = -speedScale;
          end
  
          if animationName ~= nil then
              if self.playAnimation ~= nil and self.getAnimationDuration ~= nil then
                  tipAnimation.animationName = animationName;
                  tipAnimation.animationDuration = self:getAnimationDuration(animationName);
              else
                  print("Error: tip animation "..i.." has animation name, but misses specialization AnimatedVehicle in "..self.configFileName);
              end
          else
              local animationRootNode = Utils.indexToObject(self.components, getXMLString(xmlFile, key.."#rootNode"));
              if animationRootNode ~= nil then
                  local animationCharSet = getAnimCharacterSet(animationRootNode);
                  if animationCharSet ~= nil then
                      local clip = getAnimClipIndex(animationCharSet, getXMLString(xmlFile, key.."#clip"));
                      assignAnimTrackClip(animationCharSet, 0, clip);
                      setAnimTrackLoopState(animationCharSet, 0, false);
  
                      tipAnimation.animationCharSet = animationCharSet;
                      tipAnimation.animationDuration = getAnimClipDuration(animationCharSet, clip);
                  end
              end
          end
          if tipAnimation.dischargeEndTime == nil and tipAnimation.animationDuration ~= nil then
              tipAnimation.dischargeEndTime = tipAnimation.animationDuration*2.0;
          end
          if tipAnimation.dischargeEndTime ~= nil then
              if self.isClient then
                  tipAnimation.dischargeParticleSystems = {};
                  local i = 0;
                  while true do
                      local keyPS = string.format(key..".dischargeParticleSystems.dischargeParticleSystem(%d)", i);
                      local t = getXMLString(xmlFile, keyPS .. "#type");
                      if t == nil then
                          break;
                      end;
                      local fillType = Fillable.fillTypeNameToInt[t];
                      if fillType ~= nil then
                          local currentPS = {};
                          local particleNode = Utils.loadParticleSystem(xmlFile, currentPS, keyPS, self.components, false, "$data/vehicles/particleSystems/trailerDischargeParticleSystem.i3d", self.baseDirectory);
                          tipAnimation.dischargeParticleSystems[fillType] = currentPS;
                      end;
                      i = i + 1;
                  end;
              end
  
              table.insert(self.tipAnimations, tipAnimation);
          else
              print("Error: invalid tip animation "..i.." in "..self.configFileName);
          end
          i = i + 1;
      end
  
      -- Fallback for old tipAnimation and tipDischargeEndTime
      if table.getn(self.tipAnimations) == 0 then
          local tipAnimation = {};
          tipAnimation.dischargeEndTime = getXMLFloat(xmlFile, "vehicle.tipDischargeEndTime#value");
          tipAnimation.dischargeStartTime = 0;
  
          local animationRootNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.tipAnimation#rootNode"));
          if animationRootNode ~= nil then
              local animationCharSet = getAnimCharacterSet(animationRootNode);
              if animationCharSet ~= 0 then
  
                  local clip = getAnimClipIndex(animationCharSet, getXMLString(xmlFile, "vehicle.tipAnimation#clip"));
                  assignAnimTrackClip(animationCharSet, 0, clip);
                  setAnimTrackLoopState(animationCharSet, 0, false);
  
                  tipAnimation.animationOpenSpeedScale = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.tipAnimation#speedScale"), 1);
                  tipAnimation.animationCloseSpeedScale = -tipAnimation.animationOpenSpeedScale;
                  tipAnimation.animationCharSet = animationCharSet;
                  tipAnimation.animationDuration = getAnimClipDuration(animationCharSet, clip);
                  if tipAnimation.dischargeEndTime == nil then
                      tipAnimation.dischargeEndTime = tipAnimation.animationDuration*2.0;
                  end
              end
          end
  
          if tipAnimation.dischargeEndTime ~= nil then
              if self.isClient then
                  tipAnimation.dischargeParticleSystems = {};
                  local i = 0;
                  while true do
                      local keyPS = string.format("vehicle.dischargeParticleSystems.dischargeParticleSystem(%d)", i);
                      local t = getXMLString(xmlFile, keyPS .. "#type");
                      if t == nil then
                          break;
                      end;
                      local fillType = Fillable.fillTypeNameToInt[t];
                      if fillType ~= nil then
                          local currentPS = {};
                          local particleNode = Utils.loadParticleSystem(xmlFile, currentPS, keyPS, self.components, false, "$data/vehicles/particleSystems/trailerDischargeParticleSystem.i3d", self.baseDirectory);
                          tipAnimation.dischargeParticleSystems[fillType] = currentPS;
                      end;
                      i = i + 1;
                  end;
              end
  
              table.insert(self.tipAnimations, tipAnimation);
          end
      end
      self.tipState = Trailer.TIPSTATE_CLOSED;
  
      self.fillLevelToTippedFillLevel = 1;
  
      self.tipReferencePoints = {};
      local i = 0;
      while true do
          local key = string.format("vehicle.tipReferencePoints.tipReferencePoint(%d)", i);
          if not hasXMLProperty(xmlFile, key) then
              break;
          end
  
          node = Utils.indexToObject(self.components, getXMLString(xmlFile, key.."#index"));
          if node ~= nil then
              local width = Utils.getNoNil(getXMLFloat(xmlFile, key.."#width"), 0);
              table.insert(self.tipReferencePoints, {node=node, width=width});
          end
          i = i + 1;
      end
  
      -- Fallback for old tipReferencePoint
      if table.getn(self.tipReferencePoints) == 0 then
          local tipReferencePoint = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.tipReferencePoint#index"));
          if tipReferencePoint == nil then
              tipReferencePoint = self.components[1].node;
          end
          table.insert(self.tipReferencePoints, {node=tipReferencePoint, width=0});
      end
  
      if self.isClient then
  
          self.hydraulicSoundEnabled = false;
          local hydraulicSound = getXMLString(xmlFile, "vehicle.hydraulicSound#file");
          if hydraulicSound ~= nil and hydraulicSound ~= "" then
              hydraulicSound  = Utils.getFilename(hydraulicSound, self.baseDirectory);
              self.hydraulicSound = createSample("hydraulicSound");
              loadSample(self.hydraulicSound, hydraulicSound, false);
              self.hydraulicSoundPitchOffset = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.hydraulicSound#pitchOffset"), 1.0);
              self.hydraulicSoundVolume = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.hydraulicSound#volume"), 1.0);
              setSamplePitch(self.hydraulicSound, self.hydraulicSoundPitchOffset);
          end;
  
          self.fillSoundEnabled = false;
          local fillSound = getXMLString(xmlFile, "vehicle.fillSound#file");
          if fillSound ~= nil and fillSound ~= "" then
              fillSound  = Utils.getFilename(fillSound, self.baseDirectory);
              self.fillSound = createSample("fillSound");
              loadSample(self.fillSound, fillSound, false);
              self.fillSoundPitchOffset = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.fillSound#pitchOffset"), 1.0);
              self.fillSoundVolume = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.fillSound#volume"), 1.0);
              setSamplePitch(self.fillSound, self.fillSoundPitchOffset);
          end;
  
          self.tipRotationNodes = {};
          local i = 0;
          while true do
              local key = string.format("vehicle.tipRotationNodes.tipRotationNode(%d)", i);
              if not hasXMLProperty(xmlFile, key) then
                  break;
              end;
              local node = Utils.indexToObject(self.components, getXMLString(xmlFile, key.."#index"));
              local rotSpeed = math.rad(Utils.getNoNil(getXMLFloat(xmlFile, key.."#rotSpeed"), 1)*0.001);
              if node ~= nil then
                  table.insert(self.tipRotationNodes, {node=node, rotSpeed=rotSpeed});
              end;
              i = i + 1;
          end;
  
          self.tipScrollers = {};
          local i = 0;
          while true do
              local key = string.format("vehicle.tipScrollerNodes.tipScrollerNode(%d)", i);
              if not hasXMLProperty(xmlFile, key) then
                  break;
              end;
              local shaderParameter = getXMLString(xmlFile, "vehicle.tipScrollerNodes#shaderParameter");
              local node = Utils.indexToObject(self.components, getXMLString(xmlFile, key.."#index"));
              local xSpeed, ySpeed = Utils.getVectorFromString(getXMLString(xmlFile, key.."#speed"));
              if node ~= nil then
                  table.insert(self.tipScrollers, {node=node, shaderParameter=shaderParameter, xSpeed=xSpeed, ySpeed=ySpeed});
                  setShaderParameter(node, shaderParameter, 0, 0, 0, 0, false);
              end;
              i = i + 1;
          end;
      end;
  
      local start = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.groundDropArea#startIndex"));
      local width = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.groundDropArea#widthIndex"));
      local height = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.groundDropArea#heightIndex"));
      if start ~= nil and width ~= nil and height ~= nil then
          local area = {};
          area.start = start;
          area.width = width;
          area.height = height;
          self.groundDropArea = area;
      end;
  
      self.groundAmountAccumulation = 0;
  
      self.isSelectable = true;
  
      self.allowTipDischarge = Utils.getNoNil(getXMLBool(xmlFile, "vehicle.allowTipDischarge#value"), true);
  end;
  
  function Abschieber:delete()
      if self.isClient then
          for _, tipAnimations in pairs(self.tipAnimations) do
              for _, particleSystem in pairs(tipAnimations.dischargeParticleSystems) do
                  Utils.deleteParticleSystem(particleSystem);
              end
          end
      end
  
      if self.hydraulicSound ~= nil then
          delete(self.hydraulicSound);
      end;
      if self.fillSound ~= nil then
          delete(self.fillSound);
      end;
  
  end;
  
  function Abschieber:mouseEvent(posX, posY, isDown, isUp, button)
  end;
  
  function Abschieber:keyEvent(unicode, sym, modifier, isDown)
  end;
  
  function Abschieber:readStream(streamId, connection)
      if connection:getIsServer() then
          local tipState = streamReadUIntN(streamId, 2);
          if tipState ~= Trailer.TIPSTATE_CLOSED then
              local tipReferencePointIndex = streamReadUIntN(streamId, 4);
              local animTime = streamReadFloat32(streamId);
  
              if self.currentTipReferencePointIndex ~= nil and self.currentTipReferencePointIndex ~= tipReferencePointIndex then
                  self:disableCurrentTipAnimation(0);
              end
  
              if tipState == Trailer.TIPSTATE_CLOSING then
                  self.currentTipReferencePointIndex = tipReferencePointIndex;
                  self:onEndTip(true);
              else
                  self:onStartTip(nil, tipReferencePointIndex, true);
              end;
              self:setCurrentTipAnimationTime(animTime);
          else
              self:disableCurrentTipAnimation(0);
          end;
          self.tipState = tipState;
      end;
  end;
  
  function Abschieber:wriAbschieberream(streamId, connection)
      if not connection:getIsServer() then
          assert(self.tipState >= 0 and self.tipState <= 3);
          streamWriteUIntN(streamId, self.tipState, 2);
          if self.tipState ~= Trailer.TIPSTATE_CLOSED then
              local animTime = self:getCurrentTipAnimationTime();
              streamWriteFloat32(streamId, animTime);
  
              local tipReferencePointIndex = Utils.getNoNil(self.currentTipReferencePointIndex, 1);
              assert(tipReferencePointIndex <= 15);
              streamWriteUIntN(streamId, tipReferencePointIndex, 4);
          end;
      end;
  end;
  
  --[[function Trailer:readUpdaAbschieberream(streamId, timestamp, connection)
  end;
  
  function Trailer:writeUpdaAbschieberream(streamId, connection, dirtyMask)
  end;]]
  
  function Abschieber:update(dt)
      if self:getIsActiveForInput() then
          if self ~= g_currentMission.trailerInTipRange and self:getCanTipToGround() then
              if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA) then
                  self:toggleTipState(nil);
              end;
          end;
      end;
  end;
  
  function Abschieber:updateTick(dt)
		
  
      if self.tipState == Trailer.TIPSTATE_OPENING or self.tipState == Trailer.TIPSTATE_OPEN then
  
          if self.isServer then
              self:updateTipping(dt);
			  
          end;
  
          if self.fillLevel <= 0 then
              self:onEndTip(true);
          end;
  
          if self.tipState == Trailer.TIPSTATE_OPENING then
              if self:getCurrentTipAnimationTime() >= self:getCurrentTipAnimationDuration() then
                  self.tipState = Trailer.TIPSTATE_OPEN;
  
                  if self.hydraulicSoundEnabled then
                      stopSample(self.hydraulicSound);
                      self.hydraulicSoundEnabled = false;
                  end;
              end;
          end;
  
          if self.isClient then
              if self.tipState == Trailer.TIPSTATE_OPENING then
                  if not self.hydraulicSoundEnabled and self.hydraulicSound ~= nil and self:getIsActiveForSound() then
                      playSample(self.hydraulicSound, 0, self.hydraulicSoundVolume, 0);
                      self.hydraulicSoundEnabled = true;
                  end;
              end;
  
              -- check if we still are in opening/open state. Maybe we ended tipping before.
              if (self.tipState == Trailer.TIPSTATE_OPENING or self.tipState == Trailer.TIPSTATE_OPEN) and self.fillLevel > 0 then
                  if not self.fillSoundEnabled and self.fillSound ~= nil and self:getIsActiveForSound() then
                      playSample(self.fillSound, 0, self.fillSoundVolume, 0);
                      self.fillSoundEnabled = true;
                  end;
              else
                  if self.fillSoundEnabled then
                      stopSample(self.fillSound);
                      self.fillSoundEnabled = false;
                  end;
              end;
          end;
  
      elseif self.tipState == Trailer.TIPSTATE_CLOSING then
	  
		
          
	  
			
			 if self.isClient then
          if self.currentFillPlane ~= nil then
              for _, node in ipairs(self.currentFillPlane.nodes) do
                  setVisibility(node.node, false);
              end;
              self.currentFillPlane = nil;
          end;
          if self.fillPlanes ~= nil and self.defaultFillPlane ~= nil then
              local fillPlane;
              local t = 0;
              if fillType ~= Fillable.FILLTYPE_UNKNOWN then
                  local fillTypeName = Fillable.fillTypeIntToName[fillType];
                  fillPlane = self.fillPlanes[fillTypeName];
                  t = self.fillLevel/self.capacity;
              end
              if fillPlane == nil then
                  fillPlane = self.defaultFillPlane
              end
              for _, node in ipairs(fillPlane.nodes) do
                  local x,y,z, rx,ry,rz, sx,sy,sz = node.animCurve:get(t);
  
                  setTranslation(node.node, x, y, z);
                  setRotation(node.node, rx, ry, rz);
                  setScale(node.node, sx, sy, sz);
                  setVisibility(node.node, self.fillLevel > 0 or node.alwaysVisible);
              end;
              self.currentFillPlane = fillPlane;
          end;
      end;
		
	
         
  
          if self.isClient then
              if not self.hydraulicSoundEnabled and self.hydraulicSound ~= nil and self:getIsActiveForSound() then
                  playSample(self.hydraulicSound, 0, self.hydraulicSoundVolume, 0);
                  self.hydraulicSoundEnabled = true;
              end;
          end;
  
          if self:getCurrentTipAnimationTime() <= 0.0 then
              if self.hydraulicSoundEnabled then
                  stopSample(self.hydraulicSound);
                  self.hydraulicSoundEnabled = false;
              end;
              self:disableCurrentTipAnimation();
              self.tipState = Trailer.TIPSTATE_CLOSED;
              for _, scroller in pairs(self.tipScrollers) do
                  setShaderParameter(scroller.node, scroller.shaderParameter, 0, 0, 0, 0, false);
              end;
          end;
      end;
  
      if self.isClient then
          if self.currentTipReferencePointIndex ~= nil then
              local tipAnimation = self.tipAnimations[self.currentTipReferencePointIndex];
              if tipAnimation ~= nil then
                  if (self.tipState == Trailer.TIPSTATE_OPENING or self.tipState == Trailer.TIPSTATE_OPEN) and self.fillLevel > 0 and self:getCurrentTipAnimationTime() >= tipAnimation.dischargeStartTime then
                      Utils.setEmittingState(tipAnimation.dischargeParticleSystems[self.currentFillType], true);
                  else
                      Utils.setEmittingState(tipAnimation.dischargeParticleSystems[self.currentFillType], false);
                  end;
              end
          end
  
          if self.tipState ~= Trailer.TIPSTATE_CLOSED then
              for _, node in pairs(self.tipRotationNodes) do
                  rotate(node.node, -dt*node.rotSpeed, 0, 0);
              end;
          end
      end;
  end;
  
 
  
  
  function Abschieber:updateTipping(dt)
      if not self.isServer then
          return;
      end;
      if self:getCanTip() then   
          local fillType = self.currentFillType;
  
          local fillDelta = -self.fillLevel;
          if self.currentTipReferencePointIndex ~= nil then
              local animation = self.tipAnimations[self.currentTipReferencePointIndex];
              if animation ~= nil and animation.dischargeEndTime > animation.dischargeStartTime then
                  if self:getCurrentTipAnimationTime() >= animation.dischargeStartTime then
                      local m = self.capacity/((animation.dischargeEndTime-animation.dischargeStartTime)/animation.animationOpenSpeedScale);
                      fillDelta = -m * dt;
                  else
                      fillDelta = 0;
                  end
              end
          end
  
          if fillDelta < 0 then
              local curFill = self.fillLevel;
              self:setFillLevelUpload(self.fillLevel + fillDelta, fillType);
              local fillDelta = self.fillLevel - curFill;
              self.lastFillDelta = fillDelta;
  
              if self.currentTipTrigger ~= nil then
                  if self.currentTipTrigger.updateTrailerTipping ~= nil then
                      self.currentTipTrigger:updateTrailerTipping(self, fillDelta*self.fillLevelToTippedFillLevel, fillType);
                  end
              elseif self:getCanTipToGround() then
                  if fillDelta < 0 then
                       local x, _,z = getWorldTranslation(self.groundDropArea.start);
                       local x1,_,z1 = getWorldTranslation(self.groundDropArea.width);
                       local x2,_,z2 = getWorldTranslation(self.groundDropArea.height);
  
                       local area = {x,z,x1,z1,x2,z2, -fillDelta*self.fillLevelToTippedFillLevel};
                      local area, bitType, valueAccumulation = TrailerTipAreaEvent.runLocally(area, fillType, self.groundAmountAccumulation)
                      if area ~= nil then
                          g_server:broadcastEvent(TrailerTipAreaEvent:new(area, bitType, fillType));
                      end;
                      if valueAccumulation ~= nil then
                          self.groundAmountAccumulation = valueAccumulation;
                      end;
                  end;
              end;
          end
      		
		else
          self:onEndTip();
      end;
	end;


 
  
 -- function Abschieber:getAttachedTrailersFillLevelAndCapacity(superFunc)
	-- local capacityBackup = self.capacity
	-- self.capacity = self.orgCapacity
	-- local fillLevel = 0
	-- local capacity = 0
	-- if superFunc ~= nil then
		-- fillLevel, capacity = superFunc(self)
	-- end
	-- self.capacity = capacityBackup
	-- return fillLevel, capacity
 -- end 
  
 function Abschieber:setFillLevelUpload(fillLevel, fillType, force)
		
	   if self.isClient then
          -- disable particle system if we switch the type or if the trailer is empty
          if (self.currentFillType ~= fillType or fillLevel <= 0) and self.currentTipReferencePointIndex ~= nil then
              local tipAnimation = self.tipAnimations[self.currentTipReferencePointIndex];
              if tipAnimation ~= nil then
                  Utils.setEmittingState(tipAnimation.dischargeParticleSystems[self.currentFillType], false);
              end
          end
      end
	  
       if (force == nil or force == false) and not self:allowFillType(fillType, false) then
          return
      end;
  
      self.currentFillType = fillType;
      self.fillLevel = fillLevel;
      if self.fillLevel > self.capacity then
          self.fillLevel = self.capacity;
      end;
      if self.fillLevel <= 0 then
          self.fillLevel = 0;
          self.currentFillType = Fillable.FILLTYPE_UNKNOWN;
      end;
  
  
      if self.isClient then
          if self.currentFillPlane ~= nil then
              for _, node in ipairs(self.currentFillPlane.nodes) do
                  setVisibility(node.node, false);
              end;
              self.currentFillPlane = nil;
          end;
          if self.fillPlanes ~= nil and self.defaultFillPlane ~= nil then
              local fillPlane;
              local t = 0;
              if fillType ~= Fillable.FILLTYPE_UNKNOWN then
                  local fillTypeName = Fillable.fillTypeIntToName[fillType];
                  fillPlane = self.fillPlanes[fillTypeName];
                  t = self.fillLevel/self.capacity;
              end
              if fillPlane == nil then
                  fillPlane = self.defaultFillPlane
              end
              for _, node in ipairs(fillPlane.nodes) do
			  
                  setVisibility(node.node, self.fillLevel > 0 or node.alwaysVisible);
              end;
			  
			
              self.currentFillPlane = fillPlane;
          end;
      end;
  end; 
  
  
  function Abschieber:draw()
      if self.isClient then
          if self.currentFillType ~= Fillable.FILLTYPE_UNKNOWN then
              g_currentMission:setFillTypeOverlayFillType(self.currentFillType);
          end;
  
          if self:getIsActiveForInput(true) then
  
              if self ~= g_currentMission.trailerInTipRange and self:getCanTipToGround() then
                  g_currentMission:addHelpButtonText(g_i18n:getText("Dump"), InputBinding.IMPLEMENT_EXTRA);
                  --g_currentMission:enableHudIcon("tip");
              end;
          end
      end;
  end;
  
  function Abschieber:toggleTipState(tipTrigger, tipReferencePointIndex)
  
      if self.tipState == Trailer.TIPSTATE_CLOSED or self.tipState == Trailer.TIPSTATE_CLOSING then
          self:onStartTip(tipTrigger, tipReferencePointIndex);
      else
          self:onEndTip();
      end;
  end;
  
  
  function Abschieber:onStartTip(tipTrigger, tipReferencePointIndex, noEventSend)
      -- we can not start tipping while we are tipping with another animation
      if self.isServer then
          if self.currentTipReferencePointIndex ~= nil and self.currentTipReferencePointIndex ~= tipReferencePointIndex then
              return;
          end
      end
  
      if noEventSend == nil or noEventSend == false then
          if g_server ~= nil then
              g_server:broadcastEvent(TrailerToggleTipEvent:new(self, true, tipTrigger, tipReferencePointIndex), nil, nil, self);
          else
              g_client:getServerConnection():sendEvent(TrailerToggleTipEvent:new(self, true, tipTrigger, tipReferencePointIndex));
          end;
      end;
  
      if self.isServer then
          self.currentTipTrigger = tipTrigger;
          self.groundAmountAccumulation = 0;
      end
      self:enableTipAnimation(tipReferencePointIndex, 1);
      self.currentTipReferencePointIndex = tipReferencePointIndex;
      self.tipState = Trailer.TIPSTATE_OPENING; 
  
      for _, scroller in pairs(self.tipScrollers) do
          setShaderParameter(scroller.node, scroller.shaderParameter, scroller.xSpeed, scroller.ySpeed, 0, 0, false);
      end;
  end;
  
  function Abschieber:onEndTip(noEventSend)
  
      if noEventSend == nil or noEventSend == false then
          if g_server ~= nil then
              g_server:broadcastEvent(TrailerToggleTipEvent:new(self, false), nil, nil, self);
          else
              g_client:getServerConnection():sendEvent(TrailerToggleTipEvent:new(self, false));
          end;
      end;
      if self.currentTipReferencePointIndex ~= nil then
          self:enableTipAnimation(self.currentTipReferencePointIndex, -1);
      end
  
      if self.isServer then
          self.currentTipTrigger = nil;
      end
  
      if self.fillSoundEnabled then
          stopSample(self.fillSound);
          self.fillSoundEnabled = false;
      end;
  
      self.tipState = Trailer.TIPSTATE_CLOSING;  
  end;
  
  function Abschieber:getCurrentFruitType()
      local fruitType = FruitUtil.fillTypeToFruitType[self.currentFillType];
      if fruitType == nil then
          fruitType = FruitUtil.FRUITTYPE_UNKNOWN;
      end;
      return fruitType;
  end;
  
 
  
  function Abschieber:onAttach(attacherVehicle)
  end;
  
  function Abschieber:onDetach()
      if self.deactivateOnDetach then
          Abschieber.onDeactivate(self);
      else
          Abschieber.onDeactivateSounds(self);
      end;
  end;
  
  function Abschieber:onLeave()
      if self.deactivateOnLeave then
          Abschieber.onDeactivate(self);
      else
          Abschieber.onDeactivateSounds(self);
      end;
  end;
  
  function Abschieber:onDeactivate()
      Abschieber.onDeactivateSounds(self);
  end;
  
  function Abschieber:onDeactivateSounds()
      if self.fillSoundEnabled then
          stopSample(self.fillSound);
          self.fillSoundEnabled = false;
      end;
      if self.hydraulicSoundEnabled then
          stopSample(self.hydraulicSound);
          self.hydraulicSoundEnabled = false;
      end;
  end;
  
  function Abschieber:getCanTip()
      if self.currentFillType ~= Fillable.FILLTYPE_UNKNOWN then
          if self.currentTipTrigger ~= nil then
              return g_currentMission:getIsTrailerInTipRange(self, self.currentTipTrigger, self.currentTipReferencePointIndex);
          else
              return self:getCanTipToGround();
          end
      end
      return false;
  end
  
  function Abschieber:getCanTipToGround()
      return g_currentMission.groundValueIds[self.currentFillType] ~= nil and self.groundDropArea ~= nil;
  end;
  
  function Abschieber:enableTipAnimation(tipReferencePointIndex, direction, animTime)
      if self.currentTipReferencePointIndex ~= tipReferencePointIndex then
          self:disableCurrentTipAnimation(0);
      end
  
      self.currentTipReferencePointIndex = tipReferencePointIndex;
  
      local animation = self.tipAnimations[tipReferencePointIndex];
  
      if animation ~= nil then
          local speedScale = animation.animationOpenSpeedScale;
          if direction < 0 then
              speedScale = animation.animationCloseSpeedScale;
          end
          if animation.animationName ~= nil then
              self:playAnimation(animation.animationName, speedScale, self:getAnimationTime(animation.animationName), true);
              if animTime ~= nil then
                  self:setRealAnimationTime(animation.animationName, animTime);
              end
          elseif animation.animationCharSet ~= nil then
              enableAnimTrack(animation.animationCharSet, 0);
              setAnimTrackSpeedScale(animation.animationCharSet, 0, speedScale);
              if animTime ~= nil then
                  setAnimTrackTime(animation.animationCharSet, 0, animTime);
              else
                  if direction > 0 then
                      if getAnimTrackTime(animation.animationCharSet, 0) < 0 then
                          setAnimTrackTime(animation.animationCharSet, 0, 0);
                      end;
                  else
                      if getAnimTrackTime(animation.animationCharSet, 0) > animation.animationDuration then
                          setAnimTrackTime(animation.animationCharSet, 0, animation.animationDuration);
                      end;
                  end
              end
          end
      end
  end
  
  function Abschieber:disableCurrentTipAnimation(animTime)
      if self.currentTipReferencePointIndex ~= nil then
          local animation = self.tipAnimations[self.currentTipReferencePointIndex];
          if animation ~= nil then
              Utils.setEmittingState(animation.dischargeParticleSystems[self.currentFillType], false);
  
              if animation.animationName ~= nil then
                  if animTime ~= nil then
                      self:setRealAnimationTime(animation.animationName, math.min(animTime, animation.animationDuration));
                  end
                  self:stopAnimation(animation.animationName, true); 
              elseif animation.animationCharSet ~= nil then
                  if animTime ~= nil then
                      enableAnimTrack(animation.animationCharSet, 0);
                      setAnimTrackTime(animation.animationCharSet, 0, animTime, true);
                  end
                  disableAnimTrack(animation.animationCharSet, 0); 
              end
          end
          self.currentTipReferencePointIndex = nil;
      end
  end
  
  function Abschieber:setCurrentTipAnimationTime(animTime)
      if self.currentTipReferencePointIndex ~= nil then
          local animation = self.tipAnimations[self.currentTipReferencePointIndex];
          if animation ~= nil then
              if animation.animationName ~= nil then
                  self:setRealAnimationTime(animation.animationName, math.min(animTime, animation.animationDuration));
              elseif animation.animationCharSet ~= nil then
                  setAnimTrackTime(animation.animationCharSet, 0, animTime);
              end
          end
      end
  end
  
  function Abschieber:getCurrentTipAnimationTime()
      if self.currentTipReferencePointIndex ~= nil then
          local animation = self.tipAnimations[self.currentTipReferencePointIndex];
          if animation ~= nil then
              if animation.animationName ~= nil then
                  return self:getRealAnimationTime(animation.animationName);
              elseif animation.animationCharSet ~= nil then
                  return getAnimTrackTime(animation.animationCharSet, 0);
              end
          end
      end
      return 0;
  end
  
  function Abschieber:getCurrentTipAnimationDuration()
      if self.currentTipReferencePointIndex ~= nil then
          local animation = self.tipAnimations[self.currentTipReferencePointIndex];
          if animation ~= nil then
              return animation.animationDuration;
          end
      end
      return 0;
  end